*
*   Grid for a cylindrical pipe
*
* usage: ogen [noplot] pipe -factor=<num> -order=[2/4/6/8] -interp=[e/i] -name= -xa= -xb= -radius=
* 
*   [xa,xb] : bounds on the axial length of the pipe
*   radius  : radius of the pipe
* 
* examples:
*     ogen noplot pipe -factor=1 -order=2 -interp=e 
*     ogen noplot pipe -factor=2 -order=2 -interp=e 
*     ogen noplot pipe -factor=4 -order=2 -interp=e 
*     ogen noplot pipe -factor=2 -order=2 -interp=e -xa=0. -xb=4. -name="pipe2eL4.hdf"
*     ogen noplot pipe -factor=2 -order=4 -interp=e -xa=0. -xb=4. -name="pipe2eL4.order4.hdf"
*     ogen noplot pipe -factor=3 -order=4 -interp=e -xa=0. -xb=4. -name="pipe3eL4.order4.hdf"
* 
* -- set default parameter values:
$xa=0.; $xb=2.; $radius=1.; 
* 
$order=2; $factor=1; $interp="i"; $name="";
$orderOfAccuracy = "second order"; $ng=2; $interpType = "implicit for all grids";
* 
* get command line arguments
GetOptions("order=i"=>\$order,"factor=i"=> \$factor,"interp=s"=> \$interp,"xa=f"=> \$xa,"xb=f"=> \$xb,"ya=f"=> \$radius,"name=s"=>\$name);
* 
if( $order eq 4 ){ $orderOfAccuracy="fourth order"; $ng=2; }\
elsif( $order eq 6 ){ $orderOfAccuracy="sixth order"; $ng=4; }\
elsif( $order eq 8 ){ $orderOfAccuracy="eighth order"; $ng=6; }
if( $interp eq "e" ){ $interpType = "explicit for all grids"; }
* 
$suffix = ".order$order"; 
if( $name eq "" ){ $name = "pipe" . "$interp$factor" . $suffix . ".hdf"; }
* 
* 
$ds = .1/$factor;
$pi=3.141592653;
* 
create mappings
*
$width=.5/$factor + $ds*($order-2);  # width of cylindrial grid in the radial direction
*
 box
  $delta = $width - $ds*($order-2); 
  $ya=-1.+$delta; $yb=1.-$delta; $za=$ya; $zb=$yb; 
  set corners
    $xa $xb  $ya $yb  $za $zb
  lines
   $nx = int( ($xb-$xa)/$ds + 1.5 ); $ny = int( ($yb-$ya)/$ds + 1.5 ); $nz=$ny; 
   $nx $ny $nz
  boundary conditions
     1 2 0 0 0 0
  share 
     1 2 0 0 0 0
  mappingName
     box
* pause
  exit
* ----------------
  cylinder
    orientation
      1 2 0 
    centre for cylinder
     0. 0. 0.
    bounds on the axial variable
      $xa $xb
    bounds on the radial variable
      $rb=$radius;  $ra=$rb-$width;
      $ra $rb 
    boundary conditions
      -1 -1 1 2 0 3
    share 
       0  0 1 2 0 0
    lines
      $nTheta = int( 2.*$pi*($ra+$rb)/2./$ds + 1.5 );  $nr = int( 2*$width/$ds + 1.5 ); 
      $nTheta $nx  $nr 
  mappingName
     unstretched-cylinder
* pause
  exit
*
* cluster grid lines
* 
  stretch coordinates
    Stretch r3:itanh
      $minGridSpacing = $ds/4.; 
    STP:stretch r3 itanh: position and min dx 1 $minGridSpacing
    stretch grid
    STRT:name cylinder
    exit
* 
*
*
exit
generate an overlapping grid
    box
    cylinder
  done
  change parameters
    ghost points
      all
      $ng $ng $ng $ng $ng $ng 
    order of accuracy
      $orderOfAccuracy
    interpolation type
      $interpType
    exit
*   display intermediate results
* 
  compute overlap
*  pause
  exit
*
save an overlapping grid
$name
pipe
exit
